#!/usr/bin/env python3

from functools import wraps
from copy import copy
from ramanchada2.spectrum.spectrum import Spectrum
from .dynamically_added import dynamically_added_filters
import logging

logger = logging.getLogger(__name__)


def add_spectrum_filter(fun):
    @wraps(fun)
    def retf(old_spe: Spectrum, *args, **kwargs) -> Spectrum:
        new_spe = copy(old_spe)
        new_spe._applied_processings.append(proc=fun.__name__, args=args, kwargs=kwargs)
        fun(old_spe, new_spe, *args, **kwargs)
        new_spe.write_cache()
        return new_spe
    if hasattr(Spectrum, fun.__name__):
        raise ValueError(f'redefining {fun.__name__}')
    Spectrum._available_processings.add(fun.__name__)
    setattr(Spectrum, fun.__name__, retf)
    dynamically_added_filters.add(fun.__name__)
    return retf



#!/usr/bin/env python3

import numpy as np
import numpy.typing as npt
from lmfit import Model
from lmfit import Parameters
from typing import Union
from pydantic import validate_arguments

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_theoretical_lines(
        model: Model,
        params: Parameters,
        x: Union[int, npt.NDArray[np.float64]] = 2000):
    """
    Generate spectrum from `lmfit` model.

    Parameters
    ----------
    model : lmfit.Model
        the model to be used for spectrum generation
    params : lmfit.Parameters
        the parameters to be applied to the model
    x : Union[int, npt.NDArray[np.float64]], optional
        array with x values, by default np.array(2000)
    """
    spe = Spectrum(x=x)
    spe.y = model.eval(params=params, x=spe.x)
    return spe



#!/usr/bin/env python3


"""
# Purpose
`ramanchada2` software package is meant to fill the gap between the theoretical
Raman analysis and the experimental Raman spectroscopy by providing means to
compare data of different origin. The software is in early development stage
but still able to solve practical problems.

# Features

## Read simulated data
Process simulated data by [VASP][] and [CRYSTAL][] and provide same interface.
CRYSTAL data contain intensities for multiple orientations -- laser beam
incidents perpendicularly or parallelly to the observation and information
for mono-crystals. VASP data provide data only for poly-crystals but in
different format. So the perpendicular and parallel intensities are calculated
by an implemented [algorithm][].

## Models
[LMFIT][] theoretical models can be build by spectral information obtained by
simulations or by provided by the user. These models can be fit to experimental
data, providing calibration information. At poor initial calibration the minimisation
procedure naturally fails. An iterative procedure aiming to solve this problem
was adopted in the code. On the first iteration the experimental spectrum lines
are artificially broadened. This makes it possible for the minimisation procedure
to find a parameters that are close enough to be used as an initial guess for
the second iteration. In few iterations the algorithm is able to fit to the original
experimental data. This idea is implemented and is at proof-of-concept level.

## Generate spectra
Spectra can be generated by the theoretical models. Random Poissonian noise and
artificial random-generated baseline can be added to the generated spectra, making
them convenient tools to test new methods for analysis.

## Spectrum manipulation
A number of filters can be applied to spectra (experimental and generated).
Scaling on both x and y axes is possible. Scaling could be linear or arbitrary
user defined function. A convolution is possible with set of predefined functions
as well as user defined model.

# Concept
The code is object oriented, written in python. Main elements are Spectrum and
theoretical models. Theoretical models are based on LMFIT library, while
Spectrum is a custom made class. Spectrum object contains data for x and y axes
and metadata coming from experimental files or other sources. It is planned
to add information about the uncertainties in x and y. All filters and manipulation
procedures are available as class methods. Measures are taken to preserve spectrum
instances immutable, so filters are generating new spectra, preserving the original
unchanged. Additionally, Spectrum has information about its history -- the sequence
of applied filters.

# File formats

## `.cha` vs USID/NSID
[ramanchada][] software package introduced `.cha` file format, which is an [HDF5][]
with a specific architecture. Two spectroscopy specific file formats -- [USID][] and
[NSID][] -- were considered as successors of `.cha`. USID and NSID are provided
by [pycroscopy][] package as an extension of HDF5. Several jupyter notebooks were
created in order to assess the advantages and disadvantages of such a change. These
file formats are designed to handle multidimensional spectral data, keeping track
of the modifications. These file formats does not provide big advantage in our case,
so currently `ramanchada2` supports `.cha` file format and does not support USID
or NSID.

## Cache
The concept to keep previous variants of data is employed in `ramanchada2`. If
configured so, the software saves the data for all Spectrum instances to a
tree-organized `.cha` file. When a particular chain of operations is requested
by the user, the software checks if the final result is present in the cache file,
if so it is provided, otherwise the software checks for its parent. When a parent
or some of the grand parents are present, they are taken as a starting point and
the needed steps are applied to provide the final result. The current implementation
uses [h5py][] library to access local hdf files. It is foreseen to have implementation
with [h5pyd][] that support network operations.


[CRYSTAL]: https://www.crystal.unito.it/index.php
[HDF5]: https://hdfgroup.org/
[LMFIT]: https://lmfit.github.io/lmfit-py/index.html
[NSID]: https://pycroscopy.github.io/pyNSID
[USID]: https://pycroscopy.github.io/USID
[VASP]: https://www.vasp.at/
[algorithm]: https://doi.org/10.1103/PhysRevB.54.7830
[h5py]: https://h5py.org/
[h5pyd]: https://github.com/HDFGroup/h5pyd
[pycroscopy]: https://pycroscopy.github.io/pycroscopy/
[ramanchada]: https://github.com/h2020charisma/ramanchada
"""

from __future__ import annotations

from . import spectrum
from . import theoretical_lines
__all__ = [
    'auxiliary',
    'io',
    'misc',
    'spectral_components',
    'spectrum',
    'theoretical_lines',
]
__version__ = '0.0.10'


import logging


class CustomFormatter(logging.Formatter):
    green = "\x1b[32m"
    blue = "\x1b[34m"
    yellow = "\x1b[33m"
    red = "\x1b[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    fmt = "%(asctime)s %(name)s %(levelname)s - %(message)s"
    fmt = "%(levelname)s - %(filename)s:%(lineno)d %(funcName)s() - %(message)s"

    FORMATS = {
        logging.DEBUG: green + fmt + reset,
        logging.INFO: blue + fmt + reset,
        logging.WARNING: yellow + fmt + reset,
        logging.ERROR: red + fmt + reset,
        logging.CRITICAL: bold_red + fmt + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def basicConfig(level=logging.INFO):
    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(CustomFormatter())
    logging.basicConfig(handlers=[ch], force=True)


stream = logging.StreamHandler()
stream.setFormatter(CustomFormatter())
logging.basicConfig(handlers=[stream], force=True)
logger = logging.getLogger(__name__)



#!/usr/bin/env python3
"""Create spectrum from simulation output files."""

from io import TextIOBase
from typing import Literal, Union

import numpy as np
from pydantic import validate_arguments

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor
from ramanchada2.io.simulated import read_simulated_lines


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_simulation(in_file: Union[str, TextIOBase],
                    sim_type: Literal['vasp', 'crystal_out', 'crystal_dat', 'raw_dat'],
                    use: Literal['I_tot', 'I_perp', 'I_par',
                                 'I_xx', 'I_xy', 'I_xz',
                                 'I_yy', 'I_yz', 'I_zz'] = 'I_tot'
                    ):
    """
    Generate spectrum from simulation file.

    The returned spectrum has only few x/y pairs -- one for each simulated line. Values along
    the x-axis will not be uniform. To make it uniform, one needs to resample the spectrum.
    Parameters
    ----------
    in_file_name : str or file-like object
        path to a local file, or file-like object
    sim_type : str
        'vasp': .dat file from VASP simulation
        'crystal_out': .out file from CRYSTAL simulation, not preferred
        'crystal_dat': .dat file from CRYSTAL simulation
    use: one of 'I_tot', 'I_perp', 'I_par', 'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz'
        'I_tot', 'I_perp', 'I_par': available for both CRYSTAL and VASP
        'I_xx', 'I_xy', 'I_xz', 'I_yy', 'I_yz', 'I_zz': available only for CRYSTAL
    """
    if isinstance(in_file, TextIOBase):
        labels, x, ydict = read_simulated_lines(in_file, sim_type=sim_type, use={use})
    else:
        with open(in_file) as f:
            labels, x, ydict = read_simulated_lines(f, sim_type=sim_type, use={use})
    y = ydict[use]
    spe = Spectrum(x=np.array(x), y=np.array(y))
    spe._sort_x()
    return spe



#!/usr/bin/env python3

import logging
from typing import List, Literal, Union

import numpy as np
from lmfit.models import LinearModel, lmfit_models
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import (add_spectrum_filter,
                                            add_spectrum_method)
from ramanchada2.misc.types.fit_peaks_result import FitPeaksResult
from ramanchada2.misc.types.peak_candidates import (
    ListPeakCandidateMultiModel, PeakCandidateMultiModel)

from ..spectrum import Spectrum

logger = logging.getLogger(__name__)
available_models = ['Gaussian', 'Lorentzian', 'Moffat', 'Voigt', 'PseudoVoigt', 'Pearson4', 'Pearson7']
available_models_type = Literal['Gaussian', 'Lorentzian', 'Moffat', 'Voigt', 'PseudoVoigt', 'Pearson4', 'Pearson7']


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def build_multipeak_model_params(profile: Union[available_models_type, List[available_models_type]],
                                 candidates: PeakCandidateMultiModel,
                                 baseline_model: Literal['linear', None] = 'linear',
                                 ):
    mod_list = list()
    if baseline_model == 'linear':
        mod_list.append(LinearModel(name='baseline', prefix='bl_'))
    for peak_i, peak in enumerate(candidates.peaks):
        mod_list.append(lmfit_models[profile](name=f'p{peak_i}', prefix=f'p{peak_i}_'))
    fit_model = np.sum(mod_list)
    fit_params = fit_model.make_params()
    if baseline_model == 'linear':
        fit_params['bl_slope'].set(value=candidates.base_slope, vary=False)
        fit_params['bl_intercept'].set(value=candidates.base_intercept, vary=False)

    for peak_i, peak in enumerate(candidates.peaks):
        if profile == 'Moffat':
            fwhm_factor = 2.
            height_factor = 2./peak.sigma**.5
            fit_params[f'p{peak_i}_amplitude'].set(value=peak.amplitude/height_factor)
            fit_params[f'p{peak_i}_beta'].set(value=1, min=1e-4, max=10)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma)

        elif profile == 'Voigt':
            fwhm_factor = 3.6013
            height_factor = 1/peak.sigma/2
            fit_params[f'p{peak_i}_amplitude'].set(value=peak.amplitude/height_factor)
            fit_params[f'p{peak_i}_gamma'].set(value=peak.sigma/fwhm_factor, vary=True)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma/fwhm_factor)

        elif profile == 'PseudoVoigt':
            fwhm_factor = lmfit_models[profile].fwhm_factor
            height_factor = 1/np.pi/np.sqrt(peak.sigma)/2
            fit_params[f'p{peak_i}_amplitude'].set(value=peak.amplitude/height_factor)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma/fwhm_factor)

        elif profile == 'Pearson4':
            fwhm_factor = 1
            fit_params[f'p{peak_i}_height'].set(value=peak.amplitude)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma/fwhm_factor)

        elif profile == 'Pearson7':
            fwhm_factor = 1
            height_factor = 1/2/peak.sigma
            fit_params[f'p{peak_i}_amplitude'].set(value=peak.amplitude/height_factor)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma/fwhm_factor)

        else:
            fwhm_factor = lmfit_models[profile].fwhm_factor
            height_factor = lmfit_models[profile].height_factor/peak.sigma/2
            fit_params[f'p{peak_i}_amplitude'].set(value=peak.amplitude/height_factor)
            fit_params[f'p{peak_i}_sigma'].set(value=peak.sigma)

        fit_params[f'p{peak_i}_amplitude'].set(min=0)
        fit_params[f'p{peak_i}_fwhm'].set(min=peak.fwhm*.4, max=peak.fwhm*2)
        fit_params[f'p{peak_i}_height'].set(min=peak.amplitude*.1, max=peak.amplitude*20)
        fit_params[f'p{peak_i}_center'].set(value=peak.position)

    return fit_model, fit_params


@add_spectrum_method
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def fit_peak_multimodel(spe, /, *,
                        profile: Union[available_models_type, List[available_models_type]],
                        candidates: ListPeakCandidateMultiModel,
                        no_fit=False,
                        should_break=[False],
                        kwargs_fit={},
                        vary_baseline: bool = False,
                        ) -> FitPeaksResult:
    def iter_cb(params, iter, resid, *args, **kws):
        return should_break[0]
    if no_fit:
        kwargs_fit = dict(kwargs_fit)
        kwargs_fit['max_nfev'] = 1
    fit_res = FitPeaksResult()
    for group in candidates.__root__:
        mod, par = build_multipeak_model_params(profile=profile, candidates=group)
        idx = (group.boundaries[0] < spe.x) & (spe.x < group.boundaries[1])
        x = spe.x[idx]
        y = spe.y[idx]
        for i in range(len(group.peaks)):
            par[f'p{i}_center'].set(vary=False)
        fr = mod.fit(y, x=x, params=par, iter_cb=iter_cb,  **kwargs_fit)
        for i in range(len(group.peaks)):
            par[f'p{i}_center'].set(vary=True)
        if vary_baseline:
            par['bl_slope'].set(vary=True)
            par['bl_intercept'].set(vary=True)
        fr = mod.fit(y, x=x, params=par, iter_cb=iter_cb, **kwargs_fit)
        fit_res.append(fr)
    return fit_res


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def fit_peaks_filter(
        old_spe: Spectrum,
        new_spe: Spectrum, /, *args,
        should_break=[False],
        kwargs_fit={},
        **kwargs,
        ):
    """
    Write fit result as metadata.
    """
    cand_groups = ListPeakCandidateMultiModel.validate(old_spe.result)
    new_spe.result = old_spe.fit_peak_multimodel(*args,  # type: ignore
                                                 candidates=cand_groups,
                                                 should_break=should_break,
                                                 kwargs_fit=kwargs_fit,
                                                 **kwargs).dumps()



#!/usr/bin/env python3

from typing import Union, Callable

from pydantic import validate_arguments, Field
import numpy as np
from scipy import signal, fft

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def generate_baseline(
        n_freq: int = Field(..., gt=2),
        size: int = Field(..., gt=2),
        rng_seed: Union[int, None] = None):
    rng = np.random.default_rng(rng_seed)
    k = rng.normal(0, size, size=(2, n_freq))
    k[1][0] = 0
    z = k[0] + k[1]*1j
    w = signal.windows.bohman(2*len(z))[-len(z):]
    z *= w
    z = np.concatenate([z, np.zeros(size-n_freq)])
    base = fft.irfft(z)
    base = base[:size]
    base -= base.min()
    base /= base.max()
    return base


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def add_baseline(old_spe: Spectrum, new_spe: Spectrum, /,
                 n_freq, amplitude, pedestal, func: Union[Callable, None] = None, rng_seed=None):
    """
    Add artificial baseline to the spectrum.
    A random baseline is generated in frequency domain using uniform random numbers.
    The baseline in frequency domain is tapered with bohman window to reduce the bandwidth
    of the baseline to first `n_freq` frequencies and is transformed to "time" domain.
    Additionaly by using `func` parameter the user can define arbitrary function
    to be added as baseline.

    Parameters
    ----------
    n_freq : int > 2
        number of lowest frequency bins distinct from zero
    amplitude : float
        upper boundary for the uniform random generator
    pedestal : float
        additive constant pedestal to the spectrum
    func : callable
        user defined function to be added as baseline.
        Example: func=lambda x: x*.01 + x**2*.0001
    rng_seed : int, optional
        seed for the random generator
    """
    size = len(old_spe.y)
    base = generate_baseline(n_freq=n_freq, size=size, rng_seed=rng_seed)
    y = old_spe.y + amplitude*base + pedestal
    if func is not None:
        y += func(old_spe.x) + old_spe.y
    new_spe.y = y



import numpy as np
from pydantic import validate_arguments

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import add_spectrum_method


@add_spectrum_method
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def central_moments(spe: Spectrum, /,
                    boundaries=(-np.infty, np.infty), moments=[1, 2, 3, 4], normalize=False
                    ):
    mom = dict()
    filter_idx = (spe.x >= boundaries[0]) & (spe.x < boundaries[1])
    x = spe.x[filter_idx]
    p = spe.y[filter_idx]
    p -= p.min()
    p /= p.sum()
    mom[1] = np.sum(x*p)
    mom[2] = np.sum((x - mom[1])**2 * p)
    for i in moments:
        if i <= 2:
            continue
        mom[i] = np.sum((x - mom[1])**i * p)
        if normalize and i > 2:
            mom[i] /= mom[2] ** i/2
    return [mom[i] for i in moments]



#!/usr/bin/env python3
"""Create spectrum from local files."""

import random

from ramanchada2.auxiliary.spectra.datasets2 import (get_filenames,
                                                     prepend_prefix)
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum


@add_spectrum_constructor()
def from_test_spe(index=None, **kwargs):
    """Create new spectrum from test data.

    Parameters
    ----------
    index : int, None, Optional, default is None
        if `index` is int it will be used as an index of filtered list
        if `index` is None a random spectrum will be taken
    **kwargs :
        the rest of the parameters will be used as filter
    """
    filtered = prepend_prefix(get_filenames(**kwargs))
    if index is None:
        fn = random.sample(filtered, 1)[0]
    else:
        fn = filtered[index]
    spe = Spectrum.from_local_file(fn)
    return spe



#!/usr/bin/env python3

import numpy as np
from pydantic import validate_arguments, PositiveInt

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def _moving_minimum(arr, window_size: PositiveInt):
    mov_min_left = [min(arr[max(0, i):min(i+window_size, len(arr))])
                    for i in range(len(arr))
                    ]
    mov_min_right = [min(arr[max(0, i-window_size):min(i, len(arr))])
                     for i in range(1, len(arr)+1)
                     ]
    return np.maximum.reduce([mov_min_left, mov_min_right])


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def moving_minimum(
        old_spe: Spectrum,
        new_spe: Spectrum,
        window_size: int):
    """
    Moving minimum baseline estimator.
    Successive values are calculated as minima of rolling rectangular window.

    Parameters:
    -----------
        window_size : int
    """
    new_spe.y = _moving_minimum(old_spe.y, window_size)


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def subtract_moving_minimum(
        old_spe: Spectrum,
        new_spe: Spectrum,
        window_size: int):
    new_spe.y = old_spe.y - _moving_minimum(old_spe.y, window_size)



import numpy as np
import os
import re
import pandas as pd
import logging

logger = logging.getLogger(__name__)


def read_JCAMP(file, verbose=True):
    x = []
    y = []
    # open .txt and read as lines
    with open(file) as d:
        lines = d.readlines()
    end_index = [i for i, s in enumerate(lines) if 'END' in s][-1]
    lines = lines[:end_index]
    # ## Marks meta data
    meta_lines = [ll for ll in lines if ll.startswith('##')][:-1]
    meta = dict([mm.strip('##').strip('\n').split('=') for mm in meta_lines])
    data_lines = [ll for ll in lines if not ll.startswith('##')]
    # read up to second last line
    for ll in data_lines[:-1]:
        # split line into individual numbers
        items = ll.strip('\n').strip().split()
        # 1st is x, the rest is y values
        x.append(float(items[0]))
        [y.append(float(item)) for item in items[1:]]
    # convert to np.array
    # calc num of y per x
    y_per_x = len(y)//len(x)
    x_increment = np.mean(np.diff(np.array(x))) / y_per_x
    new_x = []
    for xx in x:
        for ii in range(y_per_x):
            new_x.append(xx + x_increment*ii)
    # Read last line (may not be complete)
    items = data_lines[-1].strip('\n').strip().split()
    for ii, item in enumerate(items[1:]):
        # 1st is x, the rest is y values
        new_x.append(float(items[0]) + x_increment*ii)
        y.append(float(item))
    return np.array(new_x), np.array(y), meta


def readTXT(file, x_col=0, y_col=0, verbose=True):
    # open .txt and read as lines
    with open(file) as d:
        lines = d.readlines()
    # Find data lines and convert to np.array
    start, stop = startStop(lines)
    logger.debug("Importing " + str(stop-start+1) +
                 " data lines starting from line " + str(start) +
                 " in " + os.path.basename(file) + ".")
    data_lines = lines[start:stop]
    data = dataFromTxtLines(data_lines)
    # if columns not specified, assign x (Raman shift) and y (counts) axes
    if x_col == y_col == 0:
        # x axis is the one with mean closest to 1750
        score = 1./np.abs(data.mean(0)-1750)
        # x axis must be monotonous!
        s = np.sign(np.diff(data, axis=0))
        mono = np.array([np.all(c == c[0]) for c in s.T]) * 1.
        score *= mono
        x_col = np.argmax(score)
        # y axis is the one with maximal std/mean
        score = np.nan_to_num(data.std(0)/data.mean(0), nan=0)
        # Do not choose x axis again for y
        score[x_col] = -1000
        y_col = np.argmax(score)
        # if there's mroe than 2 columns and a header line
        if startStop(lines)[0] > 0 and data.shape[1] > 2:
            logger.debug("Found more than 2 data columns in " +
                         os.path.basename(file) + ".")
            header_line = lines[startStop(lines)[0]-1].strip('\n')
            header_line = [s.casefold() for s in re.split(';|,|\t', header_line)]
            # x axis is header line with "Shift"
            indices = [i for i, s in enumerate(header_line) if 'shift' in s]
            if indices != []:
                x_col = indices[0]
                logger.debug("X data: assigning column labelled '" +
                             header_line[x_col] + "'.")
            else:
                logger.debug("X data: assigning column # " + str(x_col) + ".")
            # y axis is header line with "Subtracted"
            indices = [i for i, s in enumerate(header_line) if 'subtracted' in s]
            if indices != []:
                y_col = indices[0]
                logger.debug("Y data: assigning column labelled '" +
                             header_line[y_col] + "'.")
            else:
                logger.debug("Y data: assigning column # " + str(y_col) + ".")
    x, y = data[:, x_col], data[:, y_col]
    # Only use unique x data points
    x, unique_ind = np.unique(x, return_index=True)
    y = y[unique_ind]
    # is x inverted?
    if all(np.diff(x) <= 0):
        x = np.flip(x)
        y = np.flip(y)
    meta_lines = lines[:startStop(lines)[0]]
    logger.debug("Importing " + str(start-1) + " metadata lines from " +
                 os.path.basename(file) + ".")
    meta_lines = [re.split(';|,|\t|=', ll.strip()) for ll in meta_lines]
    ml = {}
    for ll in meta_lines:
        ml.update({ll[0]: ll[1:]})
    # is x axis pixel numbers instead of Raman shifts?
    if all(np.diff(x) == 1) and (x[0] == 0 or x[0] == 1):
        if "Start WN" in ml:
            start_x = np.int(np.array(ml["Start WN"])[0])
        if "End WN" in ml:
            stop_x = np.int(np.array(ml["End WN"])[0])
        x = np.linspace(start_x, stop_x, len(x))
        logger.debug("X data: using linspace from " + str(start_x) + " to " +
                     str(stop_x) + " 1/cm.")
    return x, y, ml


def dataFromTxtLines(data_lines):
    data = []
    for ii, ll in enumerate(data_lines):
        ll = ll.strip('\n').replace("\t", " ")
        if ";" in ll:
            separator = ";"
        elif "," in ll and "." in ll:
            separator = ","
        elif "," in ll and " " in ll:
            separator = " "
        elif "," in ll:
            separator = ","
        else:
            separator = " "
        items = ll.split(separator)
        items = [item.replace(",", ".") for item in items]
        data.append(items)
    D = pd.DataFrame(np.array(data))
    D = D.replace(r'^\s*$', 0, regex=True)
    D = D.apply(pd.to_numeric)
    D = D.dropna()
    return D.to_numpy()


def isDataLine(line):
    line = line.strip("\n").replace("\t", " ")
    blank = all([c == " " for c in line])
    # has more than 75% digits
    digits = np.sum([d.isdigit() for d in line]) / len(line) > .25
    # apart from digits, has only ".", ";", ",", " "
    chars = all([c in '.,;+-eE ' for c in line if not c.isdigit()])
    return (not blank) & digits & chars


def startStop(lines):
    start_line, stop_line = 0, 0
    for ii, line in enumerate(lines):
        # if this is a data line and the following 5 lines are also data lines, then here is the start line
        if (len(lines) - ii) > 5 and start_line == 0:
            if all([isDataLine(ll) for ll in lines[ii:ii+5]]):
                start_line = ii
        # if this is a data line and the following 5 lines are also data lines, then here is the start line
        if (not isDataLine(line)) and stop_line <= start_line:
            stop_line = ii
    if stop_line <= start_line:
        stop_line = len(lines)-1
    return start_line, stop_line


def getYDataType(y_data):
    types = {0: "Single spectrum", 1: "Line scan", 2: "Map", 3: "Map series / volume"}
    return types[len(y_data.shape)-1]



#!/usr/bin/env python3

from __future__ import annotations
import numpy as np

from uncertainties import unumpy

from ..spectral_peak import SpectralPeak


class GaussPeak(SpectralPeak):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if {'a', 'x0', 'w'} - kwargs.keys():
            raise ValueError("'a', 'x0', 'w' arguments required")
        self.a = kwargs['a']
        self.x0 = kwargs['x0']
        self.w = kwargs['w']

    def __call__(self, x: unumpy.uarray):
        ret = 1/np.sqrt(2*np.pi)/self.w * self.a * unumpy.exp(-(x-self.x0)**2/2/self.w**2)
        return ret

    @property
    def delta(self):
        return self.x0, self.a

    @property
    def pos_amp_fwhm(self):
        return self.x0, self.a/np.sqrt(2*np.pi)/self.w, self.w*2*np.sqrt(2*np.log(2))

    @property
    def limit_3sigma(self):
        return self.x0-3*self.w, self.x0+3*self.w



#!/usr/bin/env python3

from __future__ import annotations

from ramanchada2.misc.plottable import Plottable
from ramanchada2.misc.base_class import BaseClass


class SpectralComponent(Plottable, BaseClass):
    def __init__(self, **kwargs):
        super(Plottable, self).__init__()
        super(BaseClass, self).__init__()
        self._origin = [(type(self).__name__, (), kwargs)]



#!/usr/bin/env python3

from typing import Union

import numpy as np
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def add_poisson_noise(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        scale: float = 1,
        rng_seed: Union[int, None] = None):
    r"""
    Add poisson noise to the spectrum.
    For each particular sample the noise is proportional to $\sqrt{scale*a_i}$.

    Parameters
    ----------
    scale : float, optional
        scale the amplitude of the noise, by default 1
    rng_seed : int, optional
        seed for the random generator
    """
    rng = np.random.default_rng(rng_seed)
    dat = old_spe.y + [rng.normal(0., np.sqrt(i*scale)) for i in old_spe.y]
    dat[dat < 0] = 0
    new_spe.y = np.array(dat)



import pydantic


class PositiveIntNotMultipleOf(pydantic.PositiveInt):
    not_multiple_of: int

    @classmethod
    def __get_validators__(cls):
        yield from super().__get_validators__()
        yield cls.validate_not_multiple_of

    @classmethod
    def validate_not_multiple_of(cls, value) -> int:
        if value % cls.not_multiple_of != 1:
            raise ValueError(f'Expected positive int not multiple of {cls.not_multiple_of}, got {value}')
        return value

    @classmethod
    def __modify_schema__(cls, field_schema):
        super().__modify_schema__(field_schema)
        field_schema.update({
            'not': {'multipleOf': cls.not_multiple_of}
        })


class PositiveOddInt(PositiveIntNotMultipleOf):
    not_multiple_of = 2



#!/usr/bin/env python3

from pydantic import validate_arguments

from ..spectrum import Spectrum
from ramanchada2.spectral_components.spectral_component_collection import SpectralComponentCollection
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_spectral_component_collection(
        spe_components: SpectralComponentCollection,
        x=2000):
    """
    spe_components : SpectralComponentCollection
    x : int or array-like, optional, default 2000
        x axis of the spectrum
    metadata : dict, optional
    """

    spe = Spectrum(x=x, metadata={'origin': 'generated'})  # type: ignore
    spe.y = spe_components(spe.x)
    return spe



from typing import Literal, Union
import numpy as np
import pandas as pd
from scipy.signal import wiener
from scipy import sparse
from scipy.sparse.linalg import spsolve

from pydantic import validate_arguments, PositiveInt
from ramanchada2.misc.types import PositiveOddInt

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@validate_arguments
def baseline_als(y, lam: float = 1e5, p: float = 0.001, niter: PositiveInt = 100,
                 smooth: Union[PositiveOddInt, Literal[0]] = PositiveOddInt(7)):
    if smooth > 0:
        y = wiener(y, smooth)
    L = len(y)
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    w = np.ones(L)
    for i in range(niter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        z = spsolve(Z, w*y)
        w = p * (y > z) + (1-p) * (y < z)
    return z


def baseline_snip(y0, niter: PositiveInt = 30):
    # y can't have negatives. fix by offset:
    y_offset = y0.min()
    y = y0 - y_offset
    # Spectrum must be row of a DataFrame
    raman_spectra = pd.DataFrame(y).T
    spectrum_points = len(raman_spectra.columns)
    raman_spectra_transformed = np.log(np.log(np.sqrt(raman_spectra + 1) + 1) + 1)
    working_spectra = np.zeros(raman_spectra.shape)
    for pp in np.arange(1, niter+1):
        r1 = raman_spectra_transformed.iloc[:, pp:spectrum_points - pp]
        r2 = (np.roll(raman_spectra_transformed, - pp, axis=1)[:, pp:spectrum_points - pp] +
              np.roll(raman_spectra_transformed, pp, axis=1)[:, pp:spectrum_points - pp])/2
        working_spectra = np.minimum(r1, r2)
        raman_spectra_transformed.iloc[:, pp:spectrum_points-pp] = working_spectra
    baseline = (np.exp(np.exp(raman_spectra_transformed)-1)-1)**2 - 1
    # Re-convert to np.array and apply inverse y offset to baseline
    return baseline.to_numpy()[0].T + y_offset


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def subtract_baseline_rc1_als(
        old_spe: Spectrum,
        new_spe: Spectrum,
        lam=1e5, p=0.001, niter=100, smooth=7
        ):
    new_spe.y = old_spe.y - baseline_als(old_spe.y, lam=lam, p=p, niter=niter, smooth=smooth)


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def subtract_baseline_rc1_snip(
        old_spe: Spectrum,
        new_spe: Spectrum,
        niter=30
        ):
    new_spe.y = old_spe.y - baseline_snip(old_spe.y, niter=niter)



#!/usr/bin/env python3

from typing import Union

import numpy as np
from pydantic import validate_arguments, PositiveFloat

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def add_gaussian_noise(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        sigma: PositiveFloat,
        rng_seed: Union[int, None] = None):
    r"""
    Add gaussian noise to the spectrum.
    Random number i.i.d. $N(0, \sigma)$ is added to every sample

    Parameters
    ----------
    sigma : float
        sigma of the gaussian noise
    rng_seed : int, optional
        seed for the random generator
    """
    rng = np.random.default_rng(rng_seed)
    dat = old_spe.y + rng.normal(0., sigma, size=len(old_spe.y))
    if any(dat < 0):
        dat += abs(dat.min())
    new_spe.y = np.array(dat)



#!/usr/bin/env python3

from typing import Tuple, Dict

import h5py
import logging
import numpy as np
import numpy.typing as npt
import pydantic

from ramanchada2.misc.exceptions import ChadaReadNotFoundError

logger = logging.getLogger()


@pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
def write_cha(filename: str,
              dataset: str,
              x: npt.NDArray, y: npt.NDArray, meta: Dict, h5module=None):
    data = np.stack([x, y])
    try:
        _h5 = h5module or h5py
        with _h5.File(filename, mode='a') as h5:
            if h5.get(dataset) is None:
                ds = h5.create_dataset(dataset, data=data)
                ds.attrs.update(meta)
            else:
                logger.warning(f'dataset `{dataset}` already exists in file `{filename}`')
    except ValueError as e:
        logger.warning(repr(e))


def read_cha(filename: str,
             dataset: str, h5module=None
             ) -> Tuple[npt.NDArray, npt.NDArray, Dict]:
    _h5 = h5module or h5py
    with _h5.File(filename, mode='r') as h5:
        data = h5.get(dataset)
        if data is None:
            raise ChadaReadNotFoundError(f'dataset `{dataset}` not found in file `{filename}`')
        x, y = data[:]
        meta = dict(data.attrs)
    return x, y, meta


def filter_dataset(topdomain, domain, process_file, sample=None, wavelength=None, instrument=None,
                   provider=None, investigation=None, kwargs={}, h5module=None):
    _h5 = h5module or h5py
    with _h5.File(domain) as dataset:
        if (sample is not None) and (dataset["annotation_sample"].attrs["sample"] == sample):
            process_file(topdomain, domain, **kwargs)


def visit_domain(topdomain="/", process_dataset=None, kwargs={}, h5module=None):
    _h5 = h5module or h5py
    if topdomain.endswith("/"):
        with _h5.Folder(topdomain) as domain:
            domain._getSubdomains()
            for domain in domain._subdomains:
                if domain["class"] == "folder":
                    visit_domain("{}/".format(domain["name"]), process_dataset, kwargs, h5module=_h5)
                else:
                    if not (process_dataset is None):
                        process_dataset(topdomain, domain["name"], **kwargs, h5module=_h5)
    else:
        if not (process_dataset is None):
            process_dataset(None, topdomain, **kwargs, h5module=_h5)



from abc import ABC, abstractmethod
from pydantic import BaseModel


class PydBaseModel(BaseModel, ABC):
    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def serialize(self):
        pass



#!/usr/bin/env python

from functools import wraps
from ramanchada2.spectrum.spectrum import Spectrum


def add_spectrum_method(fun):
    @wraps(fun)
    def retf(spe: Spectrum, *args, **kwargs):
        ret = fun(spe, *args, **kwargs)
        return ret
    if hasattr(Spectrum, fun.__name__):
        raise ValueError(f'redefining {fun.__name__}')
    setattr(Spectrum, fun.__name__, retf)
    return retf



from __future__ import annotations
from abc import ABC, abstractmethod
from matplotlib.axes import Axes
import matplotlib.pyplot as plt


class Plottable(ABC):
    def plot(self, ax=None, label=' ', **kwargs) -> Axes:
        if ax is None:
            fig, ax = plt.subplots(1)
        self._plot(ax, label=label, **kwargs)
        ax.legend()
        return ax

    @abstractmethod
    def _plot(self, ax, **kwargs):
        pass



#!/usr/bin/env python3

from typing import Union, Tuple, List, Literal
from scipy import signal
import numpy as np
from pydantic import (validate_arguments, PositiveInt,
                      NonNegativeFloat, NonNegativeInt
                      )

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import (add_spectrum_method,
                                            add_spectrum_filter)

from ramanchada2.misc.types.peak_candidates import ListPeakCandidateMultiModel


def peak_boundaries(spe, wlen, width, prominence):
    peaks = signal.find_peaks(spe.y, prominence=prominence, width=width, wlen=wlen)
    larr = peaks[1]['left_bases'][:]
    rarr = peaks[1]['right_bases'][:]
    lb = 0
    lbounds = list()
    rbounds = list()
    while len(larr):
        lbargmin = np.argmin(larr)
        lb = larr[lbargmin]
        rb = rarr[lbargmin]
        while True:
            group_bool = larr < rb
            if group_bool.any():
                rb = np.max(rarr[group_bool])
                rarr = rarr[~group_bool]
                larr = larr[~group_bool]
                continue
            break
        lbounds.append(lb)
        rbounds.append(rb)
    return np.array(list(zip(lbounds, rbounds)))


@add_spectrum_method
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def find_peak_multipeak(
        spe: Spectrum, /,
        prominence: Union[NonNegativeFloat, None] = None,
        wlen: Union[NonNegativeInt, None] = None,
        width: Union[int, Tuple[int, int], None] = None,
        hht_chain: Union[List[PositiveInt], None] = None,
        bgm_kwargs={},
        sharpening: Union[Literal['hht'], None] = None,
        strategy: Literal['topo', 'bayesian_gaussian_mixture', 'bgm'] = 'topo'
        ) -> ListPeakCandidateMultiModel:

    if prominence is None:
        prominence = spe.y_noise
    if not wlen:
        wlen = 200
    if width is None:
        width = 2

    if sharpening == 'hht':
        if hht_chain is None:
            hht_chain = [20]
        sharp_spe = spe.hht_sharpening_chain(movmin=hht_chain)  # type: ignore
    else:
        sharp_spe = spe

    x_arr = sharp_spe.x
    y_arr = sharp_spe.y

    def interpolate(x):
        x1 = int(x)
        x2 = x1 + 1
        y1 = x_arr[x1]
        y2 = x_arr[x2]
        return (y2-y1)/(x2-x1)*(x-x1)+y1

    boundaries = peak_boundaries(spe, prominence=prominence, width=width, wlen=wlen)
    boundaries = [(li, ri) for li, ri in boundaries if (ri-li) > 4]

    peaks, props = signal.find_peaks(y_arr,
                                     prominence=prominence,
                                     width=1,
                                     wlen=wlen)
    peak_groups = list()

    if strategy in {'bgm', 'bayesian_gaussian_mixture'}:
        bgm = sharp_spe.bayesian_gaussian_mixture(**bgm_kwargs)

        bgm_peaks = [[mean[0], np.sqrt(cov[0][0]), weight]
                     for mean, cov, weight in
                     zip(bgm.means_, bgm.covariances_, bgm.weights_)]
        bgm_peaks = sorted(bgm_peaks, key=lambda x: x[2], reverse=True)
        integral = np.sum(y_arr)
        n_peaks = (np.round(bgm.weights_, 2) > 0).sum()
        bgm_peaks = bgm_peaks[:n_peaks]

        peak_list = list()
        for mean, sigma, weight in bgm_peaks:
            peak_list.append(dict(amplitude=weight*integral*2/sigma,
                                  position=mean,
                                  sigma=sigma,
                                  ))
        for li, ri in boundaries:
            peak_group = list()
            for peak in peak_list:
                if li < peak['position'] < ri:
                    peak_group.append(dict(position=peak['position'],
                                           amplitude=peak['amplitude'],
                                           sigma=peak['sigma'])
                                      )
            if peak_group:
                peak_groups.append(dict(boundaries=(x_arr[li], x_arr[ri]),
                                        peaks=peak_group))

    elif strategy == 'topo':
        for li, ri in boundaries:
            peak_group = list()
            x1 = spe.x[li]
            x2 = spe.x[ri]
            y1 = spe.y[li]
            y2 = spe.y[ri]
            slope = (y2-y1)/(x2-x1)
            intercept = -slope*x1+y1
            for peak_i, peak_pos in enumerate(peaks):
                if li < peak_pos < ri:
                    pos_maximum = x_arr[peak_pos]
                    amplitude = props['prominences'][peak_i]
                    lwhm = pos_maximum - interpolate(props['left_ips'][peak_i])
                    rwhm = interpolate(props['right_ips'][peak_i]) - pos_maximum
                    fwhm = lwhm + rwhm
                    sigma = fwhm/2.355
                    skew = (rwhm-lwhm)/(rwhm+lwhm)
                    peak_group.append(dict(position=pos_maximum,
                                           amplitude=amplitude,
                                           sigma=sigma,
                                           skew=skew)
                                      )
            if peak_group:
                peak_groups.append(dict(base_intercept=intercept,
                                        base_slope=slope,
                                        boundaries=(x_arr[li], x_arr[ri]),
                                        peaks=peak_group))

    candidates = ListPeakCandidateMultiModel.validate(peak_groups)
    return candidates


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def find_peak_multipeak_filter(
        old_spe: Spectrum,
        new_spe: Spectrum, /,
        *args, **kwargs):
    res = old_spe.find_peak_multipeak(*args, **kwargs)  # type: ignore
    new_spe.result = res.dict()['__root__']



#!/usr/bin/env python3

import numpy as np
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_filter
from ..spectrum import Spectrum


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def pad_zeros(old_spe: Spectrum,
              new_spe: Spectrum, /):
    lenx = len(old_spe.x)
    minx = np.min(old_spe.x)
    maxx = np.max(old_spe.x)
    xl = np.linspace(minx-(maxx-minx), minx, lenx, endpoint=True)[:-1]
    xr = np.linspace(maxx, maxx+(maxx-minx), lenx, endpoint=True)[1:]

    new_spe.y = np.concatenate((np.zeros(lenx-1), old_spe.y, np.zeros(lenx-1)))
    new_spe.x = np.concatenate((xl, old_spe.x, xr))



#!/usr/bin/env python3
"""Create spectrum from local files."""

from typing import Literal, Union

from pydantic import validate_arguments

import os
from ..spectrum import Spectrum
from ramanchada2.misc.types import SpeMetadataModel
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor
from ramanchada2.io.experimental import read_txt, read_csv, rc1_parser


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_local_file(
        in_file_name: str,
        filetype: Union[None, Literal['spc', 'sp', 'spa', '0', '1', '2',
                                      'wdf', 'ngs', 'jdx', 'dx',
                                      'txt', 'txtr', 'csv', 'prn', 'rruf']] = None,
        backend: Union[None, Literal['native', 'rc1_parser']] = None):
    """
    Read experimental spectrum from a local file.

    Parameters
    ----------
    in_file_name : str
        path to a local file, containing a spectrum
    filetype : optional, default is None
        Specify the filetype. Filetype can be any of:
        `spc`, `sp`, `spa`, `0`, `1`, `2`, `wdf`, `ngs`, `jdx`,
        `dx`, `txt`, `txtr`, `csv`, `prn`, `rruf` or `None`
        `None` used to determine by extension of the file.
    backend : Literal['native', 'rc1_parser'], default None==both

    Raises
    ------
    ValueError
        When called with unsupported file formats
    """
    def load_native():
        if filetype is None:
            ft = os.path.splitext(in_file_name)[1][1:]
        else:
            ft = filetype
        if ft in {'txt', 'txtr', 'prn', 'rruf'}:
            with open(in_file_name) as fp:
                x, y, meta = read_txt(fp)
        elif ft in {'csv'}:
            with open(in_file_name) as fp:
                x, y, meta = read_csv(fp)
        else:
            raise ValueError(f'filetype {ft} not supported')
        meta["Original file"] = os.path.basename(in_file_name)
        spe = Spectrum(x=x, y=y, metadata=meta)  # type: ignore
        return spe

    def load_rc1():
        x, y, meta = rc1_parser.parse(in_file_name, filetype)
        spe = Spectrum(x=x, y=y, metadata=SpeMetadataModel.parse_obj(meta))
        return spe

    if backend == 'native':
        spe = load_native()
    elif backend == 'rc1_parser':
        spe = load_rc1()
    elif backend is None:
        try:
            spe = load_native()
        except Exception:
            spe = load_rc1()
    spe._sort_x()
    return spe



#!/usr/bin/env python3

from typing import Tuple
from scipy import signal, fft
import numpy as np

from pydantic import validate_arguments, PositiveInt

from ramanchada2.misc.spectrum_deco import add_spectrum_filter, add_spectrum_method
from ..spectrum import Spectrum


@add_spectrum_method
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def resample_NUDFT(spe: Spectrum, /,
                   x_range: Tuple[float, float] = (0, 4000),
                   xnew_bins: PositiveInt = 100,
                   window=signal.windows.hann,
                   cumulative: bool = False):
    x = spe.x
    y = spe.y
    x = np.array(x)
    x = x[(x >= x_range[0]) & (x < x_range[1])]
    x = (x-x_range[0])/(x_range[1]-x_range[0])*xnew_bins
    w = np.linspace(0, np.pi, (xnew_bins)//2+1)
    Y_new = np.sum([yi*np.exp(-1j*w*xi) for xi, yi in zip(x, y)], axis=0)
    Y_new *= (window(len(Y_new)*2))[len(Y_new):]
    y_new = fft.irfft(Y_new, n=xnew_bins)
    x_new = np.linspace(x_range[0], x_range[1], xnew_bins)
    if cumulative:
        y_new = np.cumsum(y_new)
        y_new /= y_new[-1]
    return x_new, y_new


@add_spectrum_filter
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def resample_NUDFT_filter(old_spe: Spectrum,
                          new_spe: Spectrum, /,
                          x_range: Tuple[float, float] = (0, 4000),
                          xnew_bins: PositiveInt = 100,
                          window=signal.windows.blackmanharris,
                          cumulative: bool = False):
    new_spe.x, new_spe.y = resample_NUDFT(old_spe,
                                          x_range=x_range,
                                          xnew_bins=xnew_bins,
                                          window=window,
                                          cumulative=cumulative)



#!/usr/bin/env python3

from io import TextIOBase

import pandas
from pydantic import validate_arguments


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def lines_from_vasp_dat(data_in: TextIOBase) -> pandas.DataFrame:
    """
    calculates perpendicular and parallel intensities using
    https://doi.org/10.1103/PhysRevB.54.7830
    """
    lines = data_in.readlines()
    lines_split = [ll.strip(' \r\n#').split() for ll in lines]
    df = pandas.DataFrame.from_records(data=lines_split[1:], columns=lines_split[0])
    df = df.apply(pandas.to_numeric)

    alpha = df['alpha']
    beta2 = df['beta2']
    perp_par_ratio = 3*beta2/(45*alpha**2 + 4*beta2)
    perp_par_ratio = perp_par_ratio.fillna(0)

    i_tot = df['activity']
    i_perp = i_tot * perp_par_ratio
    i_par = i_tot * (1 - perp_par_ratio)
    df = df.merge(i_par.to_frame(name='I_par'), left_index=True, right_index=True)
    df = df.merge(i_perp.to_frame(name='I_perp'), left_index=True, right_index=True)
    return df



#!/usr/bin/env python3

from __future__ import annotations

from typing import List, Tuple
from pydantic import PositiveFloat

import numpy as np

from .pydantic_base_model import PydBaseModel
from ..plottable import Plottable


class PeakModel(PydBaseModel):
    amplitude: PositiveFloat
    position: float
    sigma: PositiveFloat
    skew: float = 0

    @property
    def fwhm(self) -> float:
        return self.sigma * 2.355

    @property
    def lwhm(self) -> float:
        """
        Left width at half maximum.
        """
        return self.fwhm*(1-self.skew)/2

    @property
    def rwhm(self) -> float:
        """
        Right width at half maximum.
        """
        return self.fwhm*(1+self.skew)/2

    def serialize(self):
        self.json()


class PeakCandidateMultiModel(PydBaseModel, Plottable):
    peaks: List[PeakModel]
    base_slope: float = 0
    base_intercept: float = 0
    boundaries: Tuple[float, float]

    def plot_params_baseline(self):
        x = np.array(self.boundaries)
        return (x, self.base_slope*x + self.base_intercept)

    def plot_params_errorbar(self):
        x = self.positions
        xleft = self.lwhms
        xright = self.rwhms
        y_err = (self.amplitudes/2)
        y = y_err + self.peak_bases
        return x, y, y_err, (xleft, xright)

    @property
    def positions(self):
        return np.array([p.position for p in self.peaks])

    @property
    def sigmas(self):
        return np.array([p.sigma for p in self.peaks])

    @property
    def fwhms(self):
        return np.array([p.fwhm for p in self.peaks])

    @property
    def lwhms(self):
        return np.array([p.lwhm for p in self.peaks])

    @property
    def rwhms(self):
        return np.array([p.rwhm for p in self.peaks])

    @property
    def skews(self):
        return np.array([p.skew for p in self.peaks])

    @property
    def amplitudes(self):
        return np.array([p.amplitude for p in self.peaks])

    @property
    def bases(self):
        return self.positions * self.base_slope + self.base_intercept

    def get_pos_ampl_dict(self):
        return dict(zip(self.positions, self.amplitudes))

    def get_ampl_pos_fwhm(self):
        return np.array([
            self.amplitudes,
            self.positions,
            self.fwhms]).T

    @property
    def peak_bases(self):
        return self.positions * self.base_slope + self.base_intercept

    def _plot(self, ax, *args, label=" ", **kwargs):
        ax.errorbar(*self.plot_params_errorbar(), label=label)
        ax.plot(*self.plot_params_baseline())

    def serialize(self):
        self.json()


class ListPeakCandidateMultiModel(PydBaseModel, Plottable):
    __root__: List[PeakCandidateMultiModel]

    def get_ampl_pos_fwhm(self):
        return np.concatenate([cands.get_ampl_pos_fwhm() for cands in self.__root__])

    def get_pos_ampl_dict(self):
        return {k: v for cands in self.__root__ for k, v in cands.get_pos_ampl_dict().items()}

    def _plot(self, ax, *args, label=" ", **kwargs):
        for i, gr in enumerate(self.__root__):
            gr.plot(ax=ax, *args, label=f'{label}_{i}', **kwargs)

    def __getitem__(self, key) -> PeakCandidateMultiModel:
        return self.__root__[key]

    def __iter__(self):
        return iter(self.__root__)

    def __len__(self):
        return len(self.__root__)

    def serialize(self):
        self.json()



#!/usr/bin/env python3

from .ramanshift_to_wavelength import (abs_nm_to_shift_cm_1,
                                       shift_cm_1_to_abs_nm,
                                       abs_nm_to_shift_cm_1_dict,
                                       shift_cm_1_to_abs_nm_dict,
                                       laser_wl_nm,
                                       )

from .svd import (svd_inverse,
                  svd_solve,
                  )

from .argmin2d import (argmin2d,
                       find_closest_pairs,
                       find_closest_pairs_idx,
                       align, align_shift,
                       )



#!/usr/bin/env python3

from typing import List, Tuple

import numpy as np
from numpy.typing import NDArray


def two_column_spe(lines: List[str]) -> Tuple[NDArray, NDArray]:
    """
    Parse two column spectrum.

    Parameters
    ----------
    lines : List[str]
        list of lines to parse

    Returns
    -------
    Tuple[NDArray, NDArray]
        - positions
        - intensities
    """
    positions, intensities = np.genfromtxt(lines).T
    return positions, intensities



#!/usr/bin/env python3

from typing import Dict, Callable, Union

import numpy as np
from pydantic import validate_arguments, PositiveInt

from ..spectrum import Spectrum
from ramanchada2.misc.spectrum_deco import add_spectrum_constructor


@add_spectrum_constructor()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def from_delta_lines(
        deltas: Dict[float, float],
        xcal: Union[Callable[[float], float], None] = None,
        nbins: PositiveInt = 2000,
        **kwargs
        ):
    """
    Generate `Spectrum` with delta lines.

    Parameters
    ----------
    deltas : Dict[float, float]
        keys of the dictionary are the x positions of the deltas;
        values are the amplitudes of the corresponding deltas
    xcal : Callable, optional
        x axis calibration function
    nbins : int, optional
        Number of bins in the spectrum
    Example
    -------
    xcal=lambda x: x*3 -1000, nbins=1000
    will produce spectrum with 1000 bins in the range [-1000, 2000)
    """
    if xcal is None:
        dk = list(deltas.keys())
        dkmin, dkmax = np.min(dk), np.max(dk)
        if dkmin == dkmax:
            dkmin, dkmax = dkmin*.8, dkmax*1.2
        else:
            dkmin -= (dkmax-dkmin) * .1
            dkmax += (dkmax-dkmin) * .1
        x = np.linspace(dkmin, dkmax, nbins, endpoint=False, dtype=float)
    else:
        x = np.linspace(xcal(0), xcal(nbins), nbins, endpoint=False)
    y = np.zeros_like(x)
    for pos, ampl in deltas.items():
        idx = np.argmin(np.abs(x - pos))
        y[idx] += ampl
    spe = Spectrum(x=x, y=y, **kwargs)
    return spe



import pydantic
import numpy as np
import numpy.typing as npt
from scipy import linalg
from typing import List, Union


def argmin2d(A):
    ymin_idx = np.argmin(A, axis=0)
    xmin_idx = np.argmin(A, axis=1)
    x_idx = np.unique(xmin_idx[xmin_idx[ymin_idx[xmin_idx]] == xmin_idx])
    y_idx = np.unique(ymin_idx[ymin_idx[xmin_idx[ymin_idx]] == ymin_idx])
    matches = np.stack([y_idx, x_idx]).T
    return matches


def find_closest_pairs_idx(x, y):
    outer_dif = np.abs(np.subtract.outer(x, y))
    return argmin2d(outer_dif).T


def find_closest_pairs(x, y):
    x_idx, y_idx = find_closest_pairs_idx(x, y)
    return x[x_idx], y[y_idx]


@pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
def align(x, y,
          p0: Union[List[float], npt.NDArray] = [0, 1, 0, 0],
          func=lambda x, a0, a1, a2, a3: (a0*np.ones_like(x), a1*x, a2*x**2/1, a3*(x/1000)**3),
          max_iter: pydantic.PositiveInt = 1000):
    """
    Iteratively finds best match between x and y and evaluates the x scaling parameters.
    min((lambda(x, *p)-y)**2 | *p)
    Finds best parameters *p that minimise L2 distance between scaled x and original y
    """
    if isinstance(p0, list):
        p = np.array(p0)
    else:
        p = p0
    loss = np.infty
    cur_x = x
    for it in range(max_iter):
        cur_x = np.sum(func(x, *p), axis=0)
        x_idx, y_idx = find_closest_pairs_idx(cur_x, y)
        x_match, y_match = x[x_idx], y[y_idx]
        p_bak = p
        obj_mat = np.stack(func(x_match, *np.ones_like(p)), axis=1)
        p, *_ = linalg.lstsq(obj_mat, y_match, cond=1e-8)
        loss_bak = loss
        loss = np.sum((x_match-y_match)**2)/len(x_match)**2
        if np.allclose(p, p_bak):
            break
        if loss > loss_bak:
            pass
            return p_bak
    return p


@pydantic.validate_arguments(config=dict(arbitrary_types_allowed=True))
def align_shift(x, y,
                p0: float = 0,
                max_iter: pydantic.PositiveInt = 1000):
    loss = np.infty
    cur_x = x
    p = p0
    for it in range(max_iter):
        cur_x = x + p
        x_idx, y_idx = find_closest_pairs_idx(cur_x, y)
        x_match, y_match = x[x_idx], y[y_idx]
        p_bak = p
        p = np.mean(y_match-x_match)
        loss_bak = loss
        loss = np.sum((y_match-x_match)**2)
        if np.allclose(p, p_bak):
            break
        if loss > loss_bak:
            return p_bak
    return p



import numpy as np


def svd_solve(A, b):
    """
    Solves Ax=b
    """
    u, s, vt = np.linalg.svd(A, full_matrices=False)
    v = vt.T
    s_1 = 1/s
    s_1[np.abs(s) < 1e-8 * np.max(s)] = 0
    x = v @ (np.diag(s_1) @ (u.T@b))
    return x


def svd_inverse(mat):
    u, s, vt = np.linalg.svd(mat, full_matrices=False)
    s_1 = 1/s
    s_1[np.abs(s) < 1e-8 * np.max(s)] = 0
    return vt.T @ np.diag(s_1) @ u.T



#!/usr/bin/env python3

import datetime
import json
from typing import Any, Dict, List, Union

import numpy as np
import numpy.typing as npt
import pydantic

from ..pydantic_base_model import PydBaseModel

SpeMetadataFieldTyping = Union[
    npt.NDArray, PydBaseModel,
    pydantic.StrictBool,
    pydantic.StrictInt, float,
    datetime.datetime,
    List[Any], Dict[str, Any],
    pydantic.StrictStr]

SpeMetadataTyping = Dict[str, SpeMetadataFieldTyping]


class SpeMetadataFieldModel(PydBaseModel):
    __root__: SpeMetadataFieldTyping

    @pydantic.validator('__root__', pre=True)
    def pre_validate(cls, val):
        if isinstance(val, np.ndarray):
            return val
        if isinstance(val, str):
            if val.startswith('ramanchada2_model@'):
                # The format is:
                # ramanchada2_model@ModelName#<DATA>
                pos_at = val.index('@')
                pos_hash = val.index('#')
                model_name = val[pos_at+1:pos_hash]
                from ramanchada2.misc import types
                model = getattr(types, model_name)
                return model.validate(val[pos_hash+1:])
            if (val.startswith('[') and val.endswith(']') or
               val.startswith('{') and val.endswith('}')):
                return json.loads(val.replace("'", '"').replace(r'b"', '"'))
        return val

    def serialize(self):
        if isinstance(self.__root__, list) or isinstance(self.__root__, dict):
            return json.dumps(self.__root__)
        if isinstance(self.__root__, PydBaseModel):
            return f'ramanchada2_model@{type(self.__root__).__name__}#' + self.json()
        if isinstance(self.__root__, datetime.datetime):
            return self.__root__.isoformat()
        if isinstance(self.__root__, PydBaseModel):
            return self.__root__.serialize()
        return self.__root__


class SpeMetadataModel(PydBaseModel):
    __root__: Dict[str, SpeMetadataFieldModel]

    def __str__(self):
        return str(self.serialize())

    def serialize(self):
        return {k: v.serialize() for k, v in sorted(self.__root__.items())}

    def __getitem__(self, key: str) -> SpeMetadataFieldTyping:
        return self.__root__[key].__root__

    def _update(self, val: Dict):
        self.__root__.update(self.validate(val).__root__)

    def _del_key(self, key: str):
        del self.__root__[key]

    def _flush(self):
        self.__root__ = {}



from .baseline_base import BaseLineBase


class BaselineNumerical(BaseLineBase):
    def __init__(self, x, y):
        """
        Numerical baseline

        Parameters
        ----------
        x : array-like
        y : array-like
        """



#!/usr/bin/env python3

from .baseline import *  # noqa
from .peak_profile import *  # noqa
from .spectral_component_collection import SpectralComponentCollection  # noqa

#  raise DeprecationWarning('This module is deprecated')



#!/usr/bin/env python3

from .spectrum_filter import add_spectrum_filter
from .spectrum_constructor import add_spectrum_constructor
from .spectrum_method import add_spectrum_method
from .dynamically_added import dynamically_added_filters
from .dynamically_added import dynamically_added_constructors



from typing import List

import numpy as np
from pydantic import validate_arguments

from ramanchada2.misc.spectrum_deco import add_spectrum_constructor

from ..spectrum import Spectrum


@add_spectrum_constructor(set_applied_processing=True)
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def hdr_from_multi_exposure(spes_in: List[Spectrum]):
    """Create an HDR spectrum from several spectra with different exposures.

    The resulting spectrum will have the details in low-intensity peaks
    from long-exposure-time spectrum. As long-exposure-time
    spectrum might be sturated, the information for high-intensity
    peaks will be taken from short-exposure-time spectrum.
    This function will work on a very limited number of spectra,
    because we still do not have standardized metadata.
    """

    spes = list(sorted(spes_in, key=lambda s: float(s.meta['intigration times(ms)'])))  # type: ignore
    if not np.all([spes[0].x == s.x for s in spes]):
        raise ValueError('x-axes of the spectra should be equal')
    spes_cpms = np.array([s.y / float(s.meta['intigration times(ms)']) for s in spes])  # type: ignore
    masks = np.array(list(map(lambda s: s.y > s.meta['yaxis_max'], spes)))  # type: ignore
    y = spes_cpms[0]
    for si in range(1, len(spes_cpms)):
        y[~masks[si]] = spes_cpms[si][~masks[si]]
    return Spectrum(x=spes[0].x, y=y)
